@@ -1,7 +1,7 @@
 # ABSTRACT: take structured address data and format it according to the various global/country rules
 
 package Geo::Address::Formatter;
-$Geo::Address::Formatter::VERSION = '1.2.9';
+$Geo::Address::Formatter::VERSION = '1.3.1';
 use strict;
 use warnings;
 
@@ -10,11 +10,12 @@ use Data::Dumper;
 use File::Basename qw(dirname);
 use File::Find::Rule;
 use List::Util qw(first);
-use Mustache::Simple;
+use Text::Hogan::Compiler;
 use Try::Tiny;
 use YAML qw(Load LoadFile);
 
-my $tache = Mustache::Simple->new;
+my $THC = Text::Hogan::Compiler->new;
+my %THT_cache; # a place to store Text::Hogan::Template objects
 
 
 sub new {
@@ -109,7 +110,6 @@ sub format_address {
     my $rh_config = $self->{templates}{uc($cc)} || $self->{templates}{default};
     my $template_text = $rh_config->{address_template};
 
-    #print STDERR "t text " . Dumper $template_text;
     #print STDERR "comp " . Dumper $rh_components;
     # do we have the minimal components for an address?
     # or should we instead use the fallback template?
@@ -122,6 +122,7 @@ sub format_address {
         }
         # no fallback
     }
+    $template_text =~ s/\n/, /sg;
 
     #print STDERR "t text " . Dumper $template_text;
 
@@ -134,10 +135,17 @@ sub format_address {
     if (scalar(@$ra_unknown)){
         $rh_components->{attention} = join(', ', map { $rh_components->{$_} } @$ra_unknown);
     }
+    # warn Dumper $rh_components;
+
+
+    # get a compiled template
+    if (!defined($THT_cache{$template_text})){
+        $THT_cache{$template_text} = $THC->compile($template_text, {'numeric_string_as_string' => 1});
+    } 
+    my $compiled_template = $THT_cache{$template_text};
 
-    #warn Dumper $rh_components;
     # render it
-    my $text = $self->_render_template($template_text, $rh_components);
+    my $text = $self->_render_template($compiled_template, $rh_components);
     $text = $self->_postformat($text,$rh_config->{postformat_replace});
     $text = $self->_clean($text);
     return $text;
@@ -246,7 +254,7 @@ sub _apply_replacements {
 sub _clean {
     my $self = shift;
     my $out  = shift // '';
-    $out =~ s/[,\s]+$//;
+    $out =~ s/[\},\s]+$//;
     $out =~ s/^[,\s]+//;
 
     $out =~ s/,\s*,/, /g; # multiple commas to one   
@@ -261,28 +269,27 @@ sub _clean {
 }
 
 sub _render_template {
-    my $self             = shift;
-    my $template_content = shift;
-    my $components       = shift;
+    my $self       = shift;
+    my $THTemplate = shift;
+    my $components = shift;
 
     # Mustache calls it context
     my $context = clone($components);
-
     $context->{first} = sub {
         my $text = shift;
-        $text = $tache->render($text, $components);
-        my $selected = first { length($_) } split(/\s*\|\|\s*/, $text);
+        my $newtext = $THC->compile($text, {'numeric_string_as_string' => 1})->render($components);
+        my $selected = first { length($_) } split(/\s*\|\|\s*/, $newtext);
         return $selected;
     };
-    $template_content =~ s/\n/, /sg;
-    my $output = $tache->render($template_content, $context);
+    
+    my $output = $THTemplate->render($context);
     $output = $self->_clean($output);
 
     # is it empty?
     if ($output !~ m/\w/){
-        my $num_components = scalar(keys %$components);
-        if ($num_components == 1){  
-            foreach my $k (keys %$components){
+        my @comps = keys %$components;
+        if (scalar(@comps) == 1){  
+            foreach my $k (@comps){
                 $output = $components->{$k};
             }
         } # FIXME what if more than one?
@@ -317,7 +324,7 @@ Geo::Address::Formatter - take structured address data and format it according t
 
 =head1 VERSION
 
-version 1.2.9
+version 1.3.1
 
 =head1 SYNOPSIS
 
@@ -385,7 +392,7 @@ edf <edf@opencagedata.com>
 
 =head1 COPYRIGHT AND LICENSE
 
-This software is copyright (c) 2015 by Lokku Limited.
+This software is copyright (c) 2015 by Opencage Data Limited.
 
 This is free software; you can redistribute it and/or modify it under
 the same terms as the Perl 5 programming language system itself.
